@@ -1,4 +1,4 @@
-#!/bin/env python
+# -*- coding: utf-8 -*-
 #
 # Copyright 2013 Google Inc. All Rights Reserved.
 #
@@ -31,7 +31,7 @@ are corrupt. For this use case, it's better to create an instance of the
 BinaryPlist class and then call the Parse() method on it, with a file-like
 object as an argument.
 
-  with open("myfile.plist") as fd:
+  with open("myfile.plist", "rb") as fd:
     bplist = BinaryPlist(fd)
     top_level_object = bplist.Parse(fd)
 
@@ -56,12 +56,12 @@ Happy bplisting!
 from . import __version__
 from . import __feedback_email__
 
-import cStringIO
 import datetime
 import logging
 import math
 import os
 import plistlib
+import string
 import struct
 import xml.parsers.expat
 
@@ -82,7 +82,7 @@ class RawValue(object):
     self.value = value
 
   def __str__(self):
-    return str(self.value)
+    return self.value
 
   def __repr__(self):
     return repr(self.value)
@@ -90,6 +90,9 @@ class RawValue(object):
   def __eq__(self, other):
     return self.value == other
 
+  def __unicode__(self):
+    return u"'%s'" % ''.join([u"\\x%s" % c.encode('hex') for c in self.value])
+
 
 class CorruptReference(object):
   """Marks references to objects that are corrupt."""
@@ -141,7 +144,8 @@ class BinaryPlist(object):
   # Length of the preview we show for each object when DEBUG logging
   debug_object_preview_length = 48
 
-  def __init__(self, file_obj=None, discovery_mode=False, ultra_verbosity=True):
+  def __init__(self, file_obj=None, discovery_mode=False,
+               ultra_verbosity=False):
     """Constructor.
 
     Args:
@@ -185,14 +189,14 @@ class BinaryPlist(object):
 
   def Open(self, file_obj):
     try:
-        start_offset = file_obj.tell()
-        file_obj.seek(0, os.SEEK_END)
-        self._file_size = file_obj.tell() - start_offset
-        file_obj.seek(start_offset, os.SEEK_SET)
-        self._bplist_start_offset = start_offset
-        self.fd = file_obj
+      start_offset = file_obj.tell()
+      file_obj.seek(0, os.SEEK_END)
+      self._file_size = file_obj.tell() - start_offset
+      file_obj.seek(start_offset, os.SEEK_SET)
+      self._bplist_start_offset = start_offset
+      self.fd = file_obj
     except AttributeError:
-        raise Error("This file object doesn't support seek().")
+      raise Error("This file object doesn't support seek().")
 
   def Parse(self):
     """Parses the file descriptor at file_obj."""
@@ -371,12 +375,11 @@ class BinaryPlist(object):
     finally:
       # Remove the index from the list of traversed objects
       self.objects_traversed.remove(index)
-    if (self.debug_object_preview_length > 0
-        and len(str(obj)) > self.debug_object_preview_length):
+    output_string = ToDebugString(obj)
+    fmt = "Object %d = %s"
+    if len(output_string) > self.debug_object_preview_length:
       fmt = "Object %%d ~= %%.%ds ..." % self.debug_object_preview_length
-      self._LogUltraVerbose(fmt, index, obj)
-    else:
-      self._LogUltraVerbose("Object %d = %s", index, obj)
+    self._LogUltraVerbose(fmt, index, output_string)
     return obj
 
   def _ParseObject(self):
@@ -587,10 +590,10 @@ class BinaryPlist(object):
     return self.fd.read(strlen*char_size)
 
   def _ReadStructFromFd(self, file_obj, structure):
-    """Reads the given structucture from file_obj and returns the unpacked data.
+    """Reads the given structure from file_obj and returns the unpacked data.
 
     Raises:
-      IOError: When there wasn't enough data in file_obj to acommodate the
+      IOError: When there wasn't enough data in file_obj to accommodate the
       requested structure.
 
     Args:
@@ -673,6 +676,7 @@ class BinaryPlist(object):
     try:
       return utf16.decode("utf-16-be")
     except UnicodeDecodeError:
+      self._LogWarn("Invalid UTF-16 data")
       return RawValue(utf16)
 
   def _ParseUid(self, marker_lo):
@@ -879,7 +883,7 @@ def readPlist(pathOrFile):
     file_obj = pathOrFile
   except AttributeError:
     # Must be a path then
-    file_obj = open(pathOrFile)
+    file_obj = open(pathOrFile, "rb")
     bplist_start_offset = file_obj.tell()
 
   magicversion = file_obj.read(8)
@@ -898,3 +902,74 @@ def readPlist(pathOrFile):
       return plistlib.readPlist(file_obj)
     except xml.parsers.expat.ExpatError:
       raise FormatError("Invalid plist file.")
+
+
+def ToDebugString(string):
+  try:
+    return str(string)
+  except UnicodeEncodeError:
+    return unicode(string).encode('unicode_escape')
+
+
+
+def PlistToUnicode(o, string_encoding='safeascii', encoding_options="strict",
+                   indent=4, previous_indent=0):
+  """Returns the Unicode representation of a plist object.
+
+  This is mostly just to handle displaying unicode keys or values in
+  dictionaries as the default dict implementation escapes them, and to try to
+  represent byte strings in a more human-readable form.
+
+  Not very proud of the implementation but it is what it is.
+  """
+  if isinstance(o, dict):
+    indentation = u' '*(previous_indent+indent)
+    indentation_join = u',\n%s' % indentation
+    opening_bracket = u"{\n%s" % indentation
+    closing_bracket = u"\n%s}" % (u' '*previous_indent)
+    pieces = []
+    for k, v in o.iteritems():
+      key = PlistToUnicode(k, string_encoding=string_encoding,
+                      encoding_options=encoding_options, indent=indent,
+                      previous_indent=previous_indent+indent)
+      value = PlistToUnicode(v, string_encoding=string_encoding,
+                        encoding_options=encoding_options, indent=indent,
+                        previous_indent=previous_indent+indent)
+      pieces.append(u"%s: %s" % (key, value))
+    if not pieces:
+      return '{}'
+    return ''.join([opening_bracket,
+                    indentation_join.join(pieces),
+                    closing_bracket])
+  elif isinstance(o, str):
+    try:
+      if string_encoding == "safeascii":
+        safeascii = []
+        for c in o:
+          if c in string.printable:
+            safeascii.append(c)
+          else:
+            safeascii.append("\\x" + c.encode("hex"))
+        return u"'%s'" % (''.join(safeascii)).decode("ascii", "strict")
+      return u"'%s'" % o.decode(string_encoding, encoding_options)
+    except (UnicodeEncodeError, UnicodeDecodeError):
+      return u"'%s'" % ''.join([u"\\x%s" % c.encode('hex') for c in o])
+  elif isinstance(o, unicode):
+    # Return a quote-enclosed string
+    return u"'%s'" % o
+  elif o is NullValue:
+    return u"NULL"
+  elif o is CorruptReference:
+    return u"##CORRUPT_REFERENCE##"
+  elif o is UnknownObject:
+    return u"##UNKNOWN_OBJECT##"
+  else:
+    try:
+      return u'[%s]' % u', '.join([PlistToUnicode(piece,
+                                                  string_encoding,
+                                                  encoding_options,
+                                                  indent,
+                                                  previous_indent)
+                                   for piece in o])
+    except TypeError:
+      return unicode(o)
